
#include "main.h"
#include <stdio.h>
#include "libh523.h"

#define FLASH_KEY1                 0x45670123U
#define FLASH_KEY2                 0xCDEF89ABU
//#define FLASH_OPT_KEY1             0x08192A3BU
//#define FLASH_OPT_KEY2             0x4C5D6E7FU

#define FLASH_FLAG_SR_ERRORS (FLASH_SR_WRPERR      | FLASH_SR_PGSERR    | \
			      FLASH_SR_STRBERR     | FLASH_SR_INCERR    | \
			      FLASH_SR_OPTCHANGEERR)

void FlashClrError() {
  uint32_t u32;

  /* Check FLASH operation error flags */
	u32 = FLASH_NS->NSSR & FLASH_FLAG_SR_ERRORS;
	if (u32 != 0U) FLASH_NS->NSCCR = u32 & FLASH_FLAG_SR_ERRORS;
}

uint8_t FlashWait(uint32_t Timeout) {
  uint32_t errorflag;
  uint32_t tickstart = TimerCount_ms;

  /* Wait on BSY, WBNE and DBNE flags to be reset */
	while ((FLASH_NS->NSSR & (FLASH_SR_BSY | FLASH_SR_WBNE | FLASH_SR_DBNE)) != 0U) {
	    if ((TimerCount_ms - tickstart) >= Timeout) {
#if DEBUG_CONSOLE_ON
		printf("ERROR: FlashWait(); TimeOut=%ld\r\n", Timeout);
#endif
		return false;
	    }
	}

  /* Check FLASH operation error flags */
	errorflag = (FLASH_NS->NSSR & FLASH_FLAG_SR_ERRORS);

  /* In case of error reported in Flash SR or OPTSR registers */
	if (errorflag != 0U) {
	  /* Clear error flags */
#if DEBUG_CONSOLE_ON
	    printf("ERROR: FlashWait(); (errorflag != 0U)\r\n");
#endif
	    FLASH_NS->NSCCR = errorflag & FLASH_FLAG_SR_ERRORS;
	    return false;
	}

  /* Check FLASH End of Operation flag  */
	if ((FLASH_NS->NSSR & FLASH_SR_EOP) != 0U) {
	  /* Clear FLASH End of Operation pending bit */
	    FLASH_NS->NSCCR = FLASH_SR_EOP;
	}
	return true;
}

uint8_t FlashUnlock() {

	if (READ_BIT(FLASH->NSCR, FLASH_CR_LOCK) != 0U) {
	  /* Authorize the FLASH Control Register access */
	    WRITE_REG(FLASH->NSKEYR, FLASH_KEY1);
	    WRITE_REG(FLASH->NSKEYR, FLASH_KEY2);

	    /* Verify Flash CR is unlocked */
	    if (READ_BIT(FLASH->NSCR, FLASH_CR_LOCK) != 0U) {
#if DEBUG_CONSOLE_ON
		printf("ERROR: FlashUnlock();\r\n");
#endif
		return false;
	    }
	}
	return true;
}

uint8_t FlashLock() {

  /* Set the LOCK Bit to lock the FLASH Control Register access */
	SET_BIT(FLASH->NSCR, FLASH_CR_LOCK);

  /* Verify Flash is locked */
	if (READ_BIT(FLASH->NSCR, FLASH_CR_LOCK) == 0U) {
#if DEBUG_CONSOLE_ON
	    printf("ERROR: FlashLock();\r\n");
#endif
	    return false;
	}
	return true;
}

void FlashProgramQuadWord(uint32_t **FlashAddress, uint32_t **DataAddress) {
  uint32_t primask_bit;
  uint32_t *dest_addr = *FlashAddress;
  uint32_t *src_addr  = *DataAddress;

  /* Set PG bit */
	SET_BIT(FLASH_NS->NSCR, FLASH_CR_PG);
  /* Enter critical section: Disable interrupts to avoid any interruption during the loop */
	primask_bit = __get_PRIMASK();
	__disable_irq();

	/* Program the quad-word */
	*dest_addr++ = *src_addr++;
	*dest_addr++ = *src_addr++;
	*dest_addr++ = *src_addr++;
	*dest_addr++ = *src_addr++;

	*FlashAddress = dest_addr;
	*DataAddress = src_addr;

  /* Exit critical section: restore previous priority mask */
	__set_PRIMASK(primask_bit);
}

uint8_t FlashErase(uint32_t Sector) {

	if (Sector < 16) return false;
	if (Sector > ((FLASH_SECTOR_NB * 2)-1)) return false;
	FlashClrError();
	if (!FlashUnlock() || !FlashWait(1000)) return false;

	WRITE_REG(ICACHE->FCR, ICACHE_FCR_CBSYENDF);
	CLEAR_BIT(ICACHE->CR, ICACHE_CR_EN);

	if (Sector < FLASH_SECTOR_NB) {
    /* Reset Sector Number for Bank1 */
	    FLASH_NS->NSCR &= ~(FLASH_CR_SNB | FLASH_CR_BKSEL);
	    FLASH_NS->NSCR |= (FLASH_CR_SER | (Sector << FLASH_CR_SNB_Pos) | FLASH_CR_START);

	} else {
    /* Reset Sector Number for Bank2 */
	    Sector -= FLASH_SECTOR_NB;
	    FLASH_NS->NSCR &= ~(FLASH_CR_SNB);
	    FLASH_NS->NSCR |= (FLASH_CR_SER | FLASH_CR_BKSEL | (Sector << FLASH_CR_SNB_Pos) | FLASH_CR_START);
	}

	FlashWait(1000);
	CLEAR_BIT(FLASH_NS->NSCR, FLASH_CR_SER);
	FlashLock();

	SET_BIT(ICACHE->CR, ICACHE_CR_EN);

	return true;
}
