/**
  ******************************************************************************
  * @file    USB_Device/HID_Standalone/Src/usbd_conf.c
  * @author  MCD Application Team
  * @brief   This file implements the USB Device library callbacks and MSP
  ******************************************************************************
  * @attention
  *
  * Copyright (c) 2016 STMicroelectronics.
  * All rights reserved.
  *
  * This software is licensed under terms that can be found in the LICENSE file
  * in the root directory of this software component.
  * If no LICENSE file comes with this software, it is provided AS-IS.
  *
  ******************************************************************************
  */

#include "main.h"
#include "usbd_core.h"

/*******************************************************************************
                       LL Driver Callbacks (PCD -> USB Device Library)
*******************************************************************************/

/**
  * @brief  Reset callback.
  * @param  hpcd: PCD handle
  * @retval None
  */
void HAL_PCD_ResetCallback(PCD_HandleTypeDef *hpcd) {
// printf("HAL_PCD_ResetCallback\r\n");
  USBD_HandleTypeDef *pdev;

	pdev = hpcd->pData;
	pdev->dev_state = USBD_STATE_DEFAULT;
	pdev->ep0_state = USBD_EP0_IDLE;
	pdev->dev_config = 0U;
	pdev->dev_remote_wakeup = 0U;
	pdev->dev_test_mode = 0U;

	if ((pdev->pClass != NULL) && (pdev->pClass->DeInit != NULL)) {
	    pdev->pClass->DeInit(&USB_Dev, (uint8_t)pdev->dev_config);
	}
	/* Open EP0 OUT */
	USBD_LL_OpenEP(&USB_Dev, 0x00U, USBD_EP_TYPE_CTRL, USB_MAX_EP0_SIZE);
	pdev->ep_out[0].is_used = 1;
	pdev->ep_out[0].maxpacket = USB_MAX_EP0_SIZE;
	/* Open EP0 IN */
	USBD_LL_OpenEP(&USB_Dev, 0x80U, USBD_EP_TYPE_CTRL, USB_MAX_EP0_SIZE);
	pdev->ep_in[0].is_used = 1;
	pdev->ep_in[0].maxpacket = USB_MAX_EP0_SIZE;
//	pdev->dev_speed = USBD_SPEED_FULL;

	hpcd->USB_Address = 0;
	USB_DRD_FS->DADDR = USB_DADDR_EF;
}

void HAL_PCD_SetupStageCallback(PCD_HandleTypeDef *hpcd) {
  USBD_HandleTypeDef *pdev;

	pdev = hpcd->pData;
	USBD_ParseSetupRequest(&pdev->request, (uint8_t*)hpcd->Setup);
	pdev->ep0_state = USBD_EP0_SETUP;
	pdev->ep0_data_len = pdev->request.wLength;

// printf("HAL_PCD_SetupStageCallback %d\r\n", pdev->request.bmRequest & 0x1FU);

	switch (pdev->request.bmRequest & 0x1FU) {
	  case USB_REQ_RECIPIENT_DEVICE:
	    USBD_StdDevReq(pdev, &pdev->request);
	    break;

	  case USB_REQ_RECIPIENT_INTERFACE:
	    USBD_StdItfReq(pdev, &pdev->request);
	    break;

	  case USB_REQ_RECIPIENT_ENDPOINT:
	    USBD_StdEPReq(pdev, &pdev->request);
	    break;

	  default:
	    USBD_LL_StallEP(pdev, (pdev->request.bmRequest & 0x80U));
	    break;
	}
}

void HAL_PCD_DataOutStageCallback(PCD_HandleTypeDef *hpcd, uint8_t epnum) {
// printf("HAL_PCD_DataOutStageCallback %d\r\n", epnum);
  USBD_HandleTypeDef *pdev;
  USBD_EndpointTypeDef *pep;
  uint8_t *pdata;

	pdev = hpcd->pData;
	if (epnum == 0U) {
	  pep = &pdev->ep_out[0];
	  if (pdev->ep0_state == USBD_EP0_DATA_OUT) {
	    if (pep->rem_length > pep->maxpacket) {
	      pep->rem_length -= pep->maxpacket;
	      pdata = hpcd->OUT_ep[0].xfer_buff;
	      USBD_LL_PrepareReceive(pdev, 0, pdata, MIN(pep->rem_length, pep->maxpacket));

	    } else {
	      if (pdev->dev_state == USBD_STATE_CONFIGURED) {
		  if ((pdev->pClass != NULL) && (pdev->pClass->EP0_RxReady != NULL)) {
		      pdev->pClass->EP0_RxReady(pdev);
		  }
	      }
	      USBD_CtlSendStatus(pdev);
	    }
	  }

	} else {
	    if (pdev->dev_state == USBD_STATE_CONFIGURED) {
		if ((pdev->pClass != NULL) && (pdev->pClass->DataOut != NULL)) {
		    pdev->pClass->DataOut(pdev, epnum);
		}
	    }
	}
}

void HAL_PCD_DataInStageCallback(PCD_HandleTypeDef *hpcd, uint8_t epnum) {
// printf("HAL_PCD_DataInStageCallback %d\r\n", epnum);
  USBD_HandleTypeDef *pdev;
  USBD_EndpointTypeDef *pep;
  uint8_t *pdata;

	pdev = hpcd->pData;
	if (epnum == 0U) {
	  pep = &pdev->ep_in[0];

	  if (pdev->ep0_state == USBD_EP0_DATA_IN) {
	    if (pep->rem_length > pep->maxpacket) {
	      pep->rem_length -= pep->maxpacket;
	      pdata = hpcd->IN_ep[0].xfer_buff;
	      USBD_LL_Transmit(pdev, 0, pdata, pep->rem_length);
	      /* Prepare endpoint for premature end of transfer */
	      USBD_LL_PrepareReceive(pdev, 0U, NULL, 0U);

	    } else {
	      /* last packet is MPS multiple, so send ZLP packet */
	      if ((pep->maxpacket == pep->rem_length) &&
		  (pep->total_length >= pep->maxpacket) &&
		  (pep->total_length < pdev->ep0_data_len))
	      {
	        USBD_LL_Transmit(pdev, 0, NULL, 0);
		pdev->ep0_data_len = 0U;
		/* Prepare endpoint for premature end of transfer */
		USBD_LL_PrepareReceive(pdev, 0U, NULL, 0U);

	      } else {
		if (pdev->dev_state == USBD_STATE_CONFIGURED) {
		    if ((pdev->pClass != NULL) && (pdev->pClass->EP0_TxSent != NULL)) {
			pdev->pClass->EP0_TxSent(pdev);
		    }
		}
		USBD_LL_StallEP(pdev, 0x80U);
		USBD_CtlReceiveStatus(pdev);
	      }
	    }
	  }

	} else {
	    if (pdev->dev_state == USBD_STATE_CONFIGURED) {
		if ((pdev->pClass != NULL) && (pdev->pClass->DataIn != NULL)) {
		    pdev->pClass->DataIn(pdev, epnum);
		}
	    }
	}
}

#if USE_USB_IRQ_SOF
void HAL_PCD_SOFCallback(PCD_HandleTypeDef *hpcd) {
// printf("HAL_PCD_SOFCallback\r\n");
  const USBD_ClassTypeDef *pClass;
	if (USB_Dev.dev_state == USBD_STATE_CONFIGURED) {
	    pClass = USB_Dev.pClass;
	    if ((pClass != NULL) && (pClass->SOF != NULL)) {
		pClass->SOF(&USB_Dev);
	    }
	}
}
#endif

void HAL_PCD_SuspendCallback(PCD_HandleTypeDef *hpcd) {
// printf("HAL_PCD_SuspendCallback\r\n");
	if (USB_Dev.dev_state != USBD_STATE_SUSPENDED) {
	    USB_Dev.dev_old_state = USB_Dev.dev_state;
	    USB_Dev.dev_state = USBD_STATE_SUSPENDED;
	}
}


void HAL_PCD_ResumeCallback(PCD_HandleTypeDef *hpcd) {
// printf("HAL_PCD_ResumeCallback\r\n");
	if (USB_Dev.dev_state == USBD_STATE_SUSPENDED) {
	    USB_Dev.dev_state = USB_Dev.dev_old_state;
	}
}

/*******************************************************************************
                       LL Driver Interface (USB Device Library --> PCD)
*******************************************************************************/

/**
  * @brief  Initializes the Low Level portion of the Device driver.
  * @param  pdev: Device handle
  * @retval USBD Status
  */
USBD_StatusTypeDef USBD_LL_Init(USBD_HandleTypeDef *pdev, const USBD_DescriptorsTypeDef *pdesc) {
  int i;
  PCD_TypeDef *USBx;


	/* reset CRS registers -> release */
	RCC->APB1LRSTR |= RCC_APB1LRSTR_CRSRST;
	RCC->APB1LRSTR &= ~RCC_APB1LRSTR_CRSRST;
	RCC->CR |= RCC_CR_HSI48ON;

	/* set Prescaler Source Polarity ReloadValue */
	CRS->CFGR = CRS_CFGR_SYNCSRC_1 | (48000 - 1) |
		    (34 << CRS_CFGR_FELIM_Pos);
	/* Adjust HSI48 oscillator smooth trimming */
	MODIFY_REG(CRS->CR, CRS_CR_TRIM, (32 << CRS_CR_TRIM_Pos));
	/* Enable Automatic trimming & Frequency error counter */
	CRS->CR |= CRS_CR_AUTOTRIMEN | CRS_CR_CEN;

	/* Unlink previous class*/
//	pdev->pClass = NULL;
//	pdev->pUserData = NULL;
//	if (pdesc) pdev->pDesc = pdesc;
	pdev->dev_state = USBD_STATE_DEFAULT;
//	pdev->id = 0;

	USBx = USB_DRD_FS;
  /* Set LL Driver parameters */
	hpcd.Instance = USBx;
	hpcd.Init.dev_endpoints = MAX_USBD_EP_IO;
	hpcd.Init.phy_itface = PCD_PHY_EMBEDDED;
	hpcd.Init.speed = USBD_FS_SPEED;

  /* Link The driver to the stack */
	hpcd.pData = pdev;
	pdev->pData = &hpcd;

  /* Clear interrupt mask */
	USBx->CNTR &= ~(USB_CNTR_CTRM  | USB_CNTR_WKUPM |
			USB_CNTR_SUSPM | USB_CNTR_ERRM |
			USB_CNTR_SOFM  | USB_CNTR_ESOFM |
			USB_CNTR_RESETM | USB_CNTR_L1REQM);


  /* Initialize LL Driver */
	hpcd.State = HAL_PCD_STATE_BUSY;

	/* Configure DM DP Pins PA11 PA12*/
	MODIFY_REG(GPIOA->MODER, 
			GPIO_MODER_MODE11   | GPIO_MODER_MODE12,
			GPIO_MODER_MODE11_1 | GPIO_MODER_MODE12_1);

	MODIFY_REG(GPIOA->AFR[1], 
			GPIO_AFRH_AFSEL11 | GPIO_AFRH_AFSEL12,
			(10 << GPIO_AFRH_AFSEL11_Pos) | (10 << GPIO_AFRH_AFSEL12_Pos));


	/* Enable USB FS Clocks */
	RCC->CCIPR4 |= RCC_CCIPR4_USBSEL;
	/* Enable VDDUSB */
	PWR->USBSCR |= PWR_USBSCR_USB33SV;
	RCC->APB2ENR |= RCC_APB2ENR_USBEN;


	/*Init the Core (common init.) */
//	USB_CoreInit(USBx, hpcd.Init);
	/* Disable Host Mode  Force Device Mode */
	USBx->CNTR &= ~USB_CNTR_HOST;
	/* Force Reset IP */
	USBx->CNTR |= USB_CNTR_USBRST;
	/* Clear pending interrupts */
	USBx->ISTR = 0U;

	/* Init endpoints structures */
	for (i = 0U; i < MAX_USBD_EP_IO; i++)
	{
	  /* Init ep structure */
	  hpcd.IN_ep[i].is_in = 1U;
	  hpcd.IN_ep[i].num = i;
	  /* Control until ep is activated */
	  hpcd.IN_ep[i].type = EP_TYPE_CTRL;
//	  hpcd.IN_ep[i].maxpacket = 0U;
//	  hpcd.IN_ep[i].xfer_buff = 0U;
//	  hpcd.IN_ep[i].xfer_len = 0U;
	}

	for (i = 0U; i < MAX_USBD_EP_IO; i++)
	{
//	  hpcd.OUT_ep[i].is_in = 0U;
	  hpcd.OUT_ep[i].num = i;
	  /* Control until ep is activated */
	  hpcd.OUT_ep[i].type = EP_TYPE_CTRL;
//	  hpcd.OUT_ep[i].maxpacket = 0U;
//	  hpcd.OUT_ep[i].xfer_buff = 0U;
//	  hpcd.OUT_ep[i].xfer_len = 0U;
	}

	/* Init Device */
	/* Release Reset */
	USBx->CNTR &= ~USB_CNTR_USBRST;

  /* Set interrupt mask */
	USBx->ISTR = 0;
/*
	USBx->CNTR = USB_CNTR_CTRM  | USB_CNTR_WKUPM |
		     USB_CNTR_SUSPM | USB_CNTR_ERRM |
		     USB_CNTR_SOFM  | USB_CNTR_ESOFM |
		     USB_CNTR_RESETM | USB_CNTR_L1REQM;
*/
	/* Set USBFS Interrupt priority */
	//	HAL_NVIC_SetPriority(OTG_FS_IRQn, 5, 0);
	/* Enable USBFS Interrupt */
	NVIC_EnableIRQ(USB_DRD_FS_IRQn);

	hpcd.State= HAL_PCD_STATE_READY;

	/* USER CODE BEGIN USB_Device_Init_PreTreatment_1 */
#if 0
	HAL_PCDEx_PMAConfig(&hpcd, 0x00, PCD_SNG_BUF, 0x14);
	HAL_PCDEx_PMAConfig(&hpcd, 0x80, PCD_SNG_BUF, 0x54);
	HAL_PCDEx_PMAConfig(&hpcd, 0x81, PCD_SNG_BUF, 0x94);
	HAL_PCDEx_PMAConfig(&hpcd, 0x01, PCD_SNG_BUF, 0xD4);
	HAL_PCDEx_PMAConfig(&hpcd, 0x82, PCD_SNG_BUF, 0x114);
#else
	hpcd.OUT_ep[0].pmaadress = 0x14;
	hpcd.OUT_ep[1].pmaadress = 0xD4;

	hpcd.IN_ep[0].pmaadress = 0x54;
	hpcd.IN_ep[1].pmaadress = 0x94;
	hpcd.IN_ep[2].pmaadress = 0x114;
#endif
	return USBD_OK;
}

/**
  * @brief  Starts the Low Level portion of the Device driver. 
  * @param  pdev: Device handle
  * @retval USBD Status
  */
USBD_StatusTypeDef USBD_LL_Start(USBD_HandleTypeDef *pdev)
{
// printf("USBD_LL_Start\r\n");
#if USB_SINGLE_FS
	/* __HAL_PCD_ENABLE */
	USB_DRD_FS->ISTR = 0;
#if 1
	USB_DRD_FS->CNTR = (USE_USB_IRQ);
#else
	USB_DRD_FS->CNTR = USB_CNTR_CTRM  | USB_CNTR_WKUPM |
		     USB_CNTR_SUSPM | USB_CNTR_ERRM |
		     USB_CNTR_SOFM  | USB_CNTR_ESOFM |
		     USB_CNTR_RESETM | USB_CNTR_L1REQM;
#endif
	/* USB_DevConnect */
	USB_DRD_FS->BCDR |= USB_BCDR_DPPU;
#else
  PCD_HandleTypeDef *hpcd;
	hpcd = pdev->pData;
	__HAL_LOCK(hpcd);
	__HAL_PCD_ENABLE(hpcd);
	USB_DevConnect (hpcd->Instance);  
	__HAL_UNLOCK(hpcd);
#endif
  return USBD_OK;
}

/*
USBD_StatusTypeDef USBD_LL_Stop(USBD_HandleTypeDef *pdev)
{
// printf("USBD_LL_Stop\r\n");
  PCD_HandleTypeDef *hpcd;
	hpcd = pdev->pData;
	__HAL_LOCK(hpcd); 
	USB_DevConnect (hpcd->Instance);
	__HAL_PCD_ENABLE(hpcd);
	__HAL_UNLOCK(hpcd);
// printf("USBD_LL_Stop E\r\n");
  return USBD_OK;
}
*/

/**
  * @brief  Opens an endpoint of the Low Level Driver.
  * @param  pdev: Device handle
  * @param  ep_addr: Endpoint Number
  * @param  ep_type: Endpoint Type
  * @param  ep_mps: Endpoint Max Packet Size
  * @retval USBD Status
  */
USBD_StatusTypeDef USBD_LL_OpenEP(USBD_HandleTypeDef *pdev,
                                  uint8_t ep_addr,
                                  uint8_t ep_type,
                                  uint16_t ep_mps)
{
// printf("USBD_LL_OpenEP %d %d 0x%04x\r\n", ep_addr, ep_type, ep_mps);
  PCD_HandleTypeDef *hpcd;
  USB_EPTypeDef *ep;
  uint32_t num;

	hpcd = pdev->pData;
	num = ep_addr & MAX_USBD_EP_IO_MASK;
	if (ep_addr & 0x80) {
	    ep = &hpcd->IN_ep[num];
	} else {
	    ep = &hpcd->OUT_ep[num];
	}
	ep->maxpacket = ep_mps & 0x7FF;
	ep->type = ep_type;
	/* Set initial data PID. */
//	if (ep_type == EP_TYPE_BULK ) ep->data_pid_start = 0;

#if USB_SINGLE_FS
	USB_ActivateEndpoint(hpcd->Instance , ep);
#else
	__HAL_LOCK(hpcd);
	USB_ActivateEndpoint(hpcd->Instance , ep);
	__HAL_UNLOCK(hpcd);
#endif
  return USBD_OK;
}

/**
  * @brief  Closes an endpoint of the Low Level Driver.
  * @param  pdev: Device handle
  * @param  ep_addr: Endpoint Number
  * @retval USBD Status
  */
USBD_StatusTypeDef USBD_LL_CloseEP(USBD_HandleTypeDef *pdev, uint8_t ep_addr)
{
// printf("USBD_LL_CloseEP %d\r\n", ep_addr);
  PCD_HandleTypeDef *hpcd;
  USB_EPTypeDef *ep;
  uint32_t num;

	hpcd = pdev->pData;
	num = ep_addr & MAX_USBD_EP_IO_MASK;
	if (ep_addr & 0x80) {
	    ep = &hpcd->IN_ep[num];
	} else {
	    ep = &hpcd->OUT_ep[num];
	}
#if USB_SINGLE_FS
	USB_DeactivateEndpoint(hpcd->Instance , ep);
#else
	__HAL_LOCK(hpcd); 
	USB_DeactivateEndpoint(hpcd->Instance , ep);
	__HAL_UNLOCK(hpcd);   
#endif
// printf("USBD_LL_CloseEP E\r\n");
  return USBD_OK;
}

/**
  * @brief  Sets a Stall condition on an endpoint of the Low Level Driver.
  * @param  pdev: Device handle
  * @param  ep_addr: Endpoint Number
  * @retval USBD Status
  */
USBD_StatusTypeDef USBD_LL_StallEP(USBD_HandleTypeDef *pdev, uint8_t ep_addr)
{
// printf("USBD_LL_StallEP\r\n");
  PCD_HandleTypeDef *hpcd;
  USB_EPTypeDef *ep;
  uint32_t num;

	hpcd = pdev->pData;
	num = ep_addr & MAX_USBD_EP_IO_MASK;
#if USB_SINGLE_FS
	if (ep_addr & 0x80) {
	    ep = &hpcd->IN_ep[num];
	    ep->is_stall = 1;
	    USB_DRD_SET_CHEP_TX_STATUS(hpcd->Instance, num, USB_EP_TX_STALL);
	} else {
	    ep = &hpcd->OUT_ep[num];
	    ep->is_stall = 1;
	    USB_DRD_SET_CHEP_RX_STATUS(hpcd->Instance, num, USB_EP_RX_STALL);
	}
#else
	if (ep_addr & 0x80) {
	    ep = &hpcd->IN_ep[num];
	} else {
	    ep = &hpcd->OUT_ep[num];
	}
	ep->is_stall = 1;

	__HAL_LOCK(hpcd); 
	USB_EPSetStall(hpcd->Instance , ep);
	__HAL_UNLOCK(hpcd);
#endif
// printf("USBD_LL_StallEP E\r\n");
  return USBD_OK;
}

/**
  * @brief  Clears a Stall condition on an endpoint of the Low Level Driver.
  * @param  pdev: Device handle
  * @param  ep_addr: Endpoint Number
  * @retval USBD Status
  */
USBD_StatusTypeDef USBD_LL_ClearStallEP(USBD_HandleTypeDef *pdev, uint8_t ep_addr)
{
// printf("USBD_LL_ClearStallEP\r\n");
  PCD_HandleTypeDef *hpcd;
  USB_EPTypeDef *ep;
  uint32_t num;

	hpcd = pdev->pData;
	num = ep_addr & MAX_USBD_EP_IO_MASK;
#if USB_SINGLE_FS
	if (ep_addr & 0x80) {
	    ep = &hpcd->IN_ep[num];
	    ep->is_stall = 0;
	    USB_DRD_CLEAR_TX_DTOG(hpcd->Instance, num);
	    if (ep->type != EP_TYPE_ISOC) {
	      /* Configure NAK status for the Endpoint */
		USB_DRD_SET_CHEP_TX_STATUS(hpcd->Instance, num, USB_EP_TX_NAK);
	    }
	} else {
	    ep = &hpcd->OUT_ep[num];
	    ep->is_stall = 0;
	    USB_DRD_CLEAR_RX_DTOG(hpcd->Instance, num);
	    /* Configure VALID status for the Endpoint */
	    USB_DRD_SET_CHEP_RX_STATUS(hpcd->Instance, num, USB_EP_RX_VALID);
	}
#else
	if (ep_addr & 0x80) {
	    ep = &hpcd->IN_ep[num];
	} else {
	    ep = &hpcd->OUT_ep[num];
	}
	ep->is_stall = 0;

	__HAL_LOCK(hpcd);
	USB_EPClearStall(hpcd->Instance , ep);
	__HAL_UNLOCK(hpcd);
#endif
// printf("USBD_LL_ClearStallEP E\r\n");
	return USBD_OK;
}

/**
  * @brief  Returns Stall condition.
  * @param  pdev: Device handle
  * @param  ep_addr: Endpoint Number
  * @retval Stall (1: Yes, 0: No)
  */
uint8_t USBD_LL_IsStallEP(USBD_HandleTypeDef *pdev, uint8_t ep_addr) {
// printf("USBD_LL_IsStallEP\r\n");
  PCD_HandleTypeDef *hpcd;
  USB_EPTypeDef *ep;
  uint32_t num;

	hpcd = pdev->pData;
	num = ep_addr & MAX_USBD_EP_IO_MASK;
	if (ep_addr & 0x80) {
	    ep = &hpcd->IN_ep[num];
	} else {
	    ep = &hpcd->OUT_ep[num];
	}
	return ep->is_stall;
}

#if 0
/**
  * @brief  Assigns a USB address to the device.
  * @param  pdev: Device handle
  * @param  ep_addr: Endpoint Number
  * @retval USBD Status
  */
USBD_StatusTypeDef USBD_LL_SetUSBAddress(USBD_HandleTypeDef *pdev, uint8_t dev_addr)
{
// printf("USBD_LL_SetUSBAddress\r\n");
  PCD_HandleTypeDef *hpcd;
	hpcd = pdev->pData;
#if USB_SINGLE_FS
	hpcd->USB_Address = dev_addr;
	USB_DRD_FS->DADDR = USB_DADDR_EF;
#else
	__HAL_LOCK(hpcd);
	hpcd->USB_Address = dev_addr;
	(void)USB_SetDevAddress(hpcd->Instance, dev_addr);
	__HAL_UNLOCK(hpcd);
#endif
// printf("USBD_LL_SetUSBAddress E\r\n");
  return USBD_OK; 
}
#endif

/**
  * @brief  Transmits data over an endpoint.
  * @param  pdev: Device handle
  * @param  ep_addr: Endpoint Number
  * @param  pbuf: Pointer to data to be sent
  * @param  size: Data size    
  * @retval USBD Status
  */
USBD_StatusTypeDef USBD_LL_Transmit(USBD_HandleTypeDef *pdev, 
                                    uint8_t ep_addr,
                                    uint8_t *pbuf,
                                    uint32_t size)
{
  /* Get the packet total length */
// printf("USBD_LL_Transmit %d\r\n", (int)size);
  PCD_HandleTypeDef *hpcd;
  USB_EPTypeDef *ep;

	pdev->ep_in[ep_addr & MAX_USBD_EP_IO_MASK].total_length = size;
	hpcd = pdev->pData;
	ep = &hpcd->IN_ep[ep_addr & MAX_USBD_EP_IO_MASK];

	/*setup and start the Xfer */
	ep->xfer_buff = pbuf;
	ep->xfer_len = size;

	ep->xfer_count = 0U;
#if 1
	USB_EPStartXfer(hpcd->Instance, ep);
#else
  uint32_t len;
	if (ep->xfer_len > USB_FS_MAX_PACKET_SIZE) {
	  len = USB_FS_MAX_PACKET_SIZE;
	} else {
	  len = ep->xfer_len;
	}
	USB_WritePMA(hpcd->Instance, ep->xfer_buff, ep->pmaadress, (uint16_t)len);
	USB_DRD_SET_CHEP_TX_CNT(hpcd->Instance, ep->num, len);
	USB_DRD_SET_CHEP_TX_STATUS(hpcd->Instance, ep->num, USB_EP_TX_VALID);
#endif
	return USBD_OK;
}

/**
  * @brief  Prepares an endpoint for reception.
  * @param  pdev: Device handle
  * @param  ep_addr: Endpoint Number
  * @param  pbuf: Pointer to data to be received
  * @param  size: Data size
  * @retval USBD Status
  */
USBD_StatusTypeDef USBD_LL_PrepareReceive(USBD_HandleTypeDef *pdev, 
                                          uint8_t ep_addr,
                                          uint8_t *pbuf,
                                          uint32_t size)
{
// printf("USBD_LL_PrepareReceive %d %d\r\n", ep_addr, (int)size);
  PCD_HandleTypeDef *hpcd;
  USB_EPTypeDef *ep;

	hpcd = pdev->pData;
	ep = &hpcd->OUT_ep[ep_addr & MAX_USBD_EP_IO_MASK];
	ep->xfer_buff = pbuf;  
	ep->xfer_len = size;
	ep->xfer_count = 0;
	USB_EPStartXfer (hpcd->Instance, ep);
	return USBD_OK;
}

/**
  * @brief  Returns the last transferred packet size.
  * @param  pdev: Device handle
  * @param  ep_addr: Endpoint Number
  * @retval Received Data Size
  */
uint32_t USBD_LL_GetRxDataSize(USBD_HandleTypeDef *pdev, uint8_t ep_addr) {
  uint32_t result;
  PCD_HandleTypeDef *hpcd;

	hpcd = pdev->pData;
	result = hpcd->OUT_ep[ep_addr & MAX_USBD_EP_IO_MASK].xfer_count;
// printf("USBD_LL_GetRxDataSize %d\r\n", (int)result);
	return result;
}
