/**
  ******************************************************************************
  * @file    usbd_req.c
  * @author  MCD Application Team
  * @brief   This file provides the standard USB requests following chapter 9.
  ******************************************************************************
  * @attention
  *
  * Copyright (c) 2015 STMicroelectronics.
  * All rights reserved.
  *
  * This software is licensed under terms that can be found in the LICENSE file
  * in the root directory of this software component.
  * If no LICENSE file comes with this software, it is provided AS-IS.
  *
  ******************************************************************************
  */

/* Includes ------------------------------------------------------------------*/
//#include "usb_desc.h"
#include "usbd_ctlreq.h"
#include "usbd_core.h"

/** @defgroup USBD_REQ_Private_Defines
  * @{
  */
#ifndef USBD_MAX_STR_DESC_SIZ
#define USBD_MAX_STR_DESC_SIZ                          64U
#endif /* USBD_MAX_STR_DESC_SIZ */

/** @defgroup USBD_REQ_Private_FunctionPrototypes
  * @{
  */
static void USBD_GetDescriptor(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req);
static void USBD_SetAddress(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req);
static USBD_StatusTypeDef USBD_SetConfig(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req);
static void USBD_GetConfig(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req);
static void USBD_GetStatus(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req);
static void USBD_SetFeature(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req);
static void USBD_ClrFeature(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req);
static uint8_t USBD_GetLen(uint8_t *buf);

/**
  * @}
  */


/** @defgroup USBD_REQ_Private_Functions
  * @{
  */

/**
  * @brief  USBD_CtlSendStatus
  *         send zero lzngth packet on the ctl pipe
  * @param  pdev: device instance
  * @retval status
  */
USBD_StatusTypeDef USBD_CtlSendStatus(USBD_HandleTypeDef *pdev)
{
  /* Set EP0 State */
  pdev->ep0_state = USBD_EP0_STATUS_IN;

  /* Start the transfer */
  (void)USBD_LL_Transmit(pdev, 0x00U, NULL, 0U);

  return USBD_OK;
}

/**
  * @brief  USBD_CtlReceiveStatus
  *         receive zero lzngth packet on the ctl pipe
  * @param  pdev: device instance
  * @retval status
  */
USBD_StatusTypeDef USBD_CtlReceiveStatus(USBD_HandleTypeDef *pdev)
{
  /* Set EP0 State */
  pdev->ep0_state = USBD_EP0_STATUS_OUT;

  /* Start the transfer */
  (void)USBD_LL_PrepareReceive(pdev, 0U, NULL, 0U);

  return USBD_OK;
}

/**
  * @brief  USBD_CtlSendData
  *         send data on the ctl pipe
  * @param  pdev: device instance
  * @param  buff: pointer to data buffer
  * @param  len: length of data to be sent
  * @retval status
  */
USBD_StatusTypeDef USBD_CtlSendData(USBD_HandleTypeDef *pdev,
                                    uint8_t *pbuf, uint32_t len)
{
  /* Set EP0 State */
  pdev->ep0_state = USBD_EP0_DATA_IN;
  pdev->ep_in[0].total_length = len;

#ifdef USBD_AVOID_PACKET_SPLIT_MPS
  pdev->ep_in[0].rem_length = 0U;
#else
  pdev->ep_in[0].rem_length = len;
#endif /* USBD_AVOID_PACKET_SPLIT_MPS */

  /* Start the transfer */
  (void)USBD_LL_Transmit(pdev, 0x00U, pbuf, len);

  return USBD_OK;
}

/**
  * @brief  USBD_CtlPrepareRx
  *         receive data on the ctl pipe
  * @param  pdev: device instance
  * @param  buff: pointer to data buffer
  * @param  len: length of data to be received
  * @retval status
  */
USBD_StatusTypeDef USBD_CtlPrepareRx(USBD_HandleTypeDef *pdev,
                                     uint8_t *pbuf, uint32_t len)
{
  /* Set EP0 State */
  pdev->ep0_state = USBD_EP0_DATA_OUT;
  pdev->ep_out[0].total_length = len;

#ifdef USBD_AVOID_PACKET_SPLIT_MPS
  pdev->ep_out[0].rem_length = 0U;
#else
  pdev->ep_out[0].rem_length = len;
#endif /* USBD_AVOID_PACKET_SPLIT_MPS */

  /* Start the transfer */
  (void)USBD_LL_PrepareReceive(pdev, 0U, pbuf, len);

  return USBD_OK;
}

/**
  * @brief  USBD_StdDevReq
  *         Handle standard usb device requests
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval status
  */
USBD_StatusTypeDef USBD_StdDevReq(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
  USBD_StatusTypeDef ret = USBD_OK;

  switch (req->bmRequest & USB_REQ_TYPE_MASK)
  {
    case USB_REQ_TYPE_CLASS:
    case USB_REQ_TYPE_VENDOR:
      ret = (USBD_StatusTypeDef)pdev->pClass->Setup(pdev, req);
      break;

    case USB_REQ_TYPE_STANDARD:
      switch (req->bRequest)
      {
        case USB_REQ_GET_DESCRIPTOR:
          USBD_GetDescriptor(pdev, req);
          break;

        case USB_REQ_SET_ADDRESS:
          USBD_SetAddress(pdev, req);
          break;

        case USB_REQ_SET_CONFIGURATION:
          ret = USBD_SetConfig(pdev, req);
          break;

        case USB_REQ_GET_CONFIGURATION:
          USBD_GetConfig(pdev, req);
          break;

        case USB_REQ_GET_STATUS:
          USBD_GetStatus(pdev, req);
          break;

        case USB_REQ_SET_FEATURE:
          USBD_SetFeature(pdev, req);
          break;

        case USB_REQ_CLEAR_FEATURE:
          USBD_ClrFeature(pdev, req);
          break;

        default:
          USBD_CtlError(pdev, req);
          break;
      }
      break;

    default:
      USBD_CtlError(pdev, req);
      break;
  }

  return ret;
}

/**
  * @brief  USBD_StdItfReq
  *         Handle standard usb interface requests
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval status
  */
USBD_StatusTypeDef USBD_StdItfReq(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
  USBD_StatusTypeDef ret = USBD_OK;

  switch (req->bmRequest & USB_REQ_TYPE_MASK) {
    case USB_REQ_TYPE_CLASS:
    case USB_REQ_TYPE_VENDOR:
    case USB_REQ_TYPE_STANDARD:
      switch (pdev->dev_state) {
        case USBD_STATE_DEFAULT:
        case USBD_STATE_ADDRESSED:
        case USBD_STATE_CONFIGURED:

          if (LOBYTE(req->wIndex) <= USBD_MAX_NUM_INTERFACES) {
	    if (pdev->pClass->Setup != NULL) {
	      pdev->classId = 0;
	      ret = (USBD_StatusTypeDef)(pdev->pClass->Setup(pdev, req));
	    }
            if ((req->wLength == 0U) && (ret == USBD_OK)) {
              (void)USBD_CtlSendStatus(pdev);
            }
          } else {
            USBD_CtlError(pdev, req);
          }
          break;

        default:
          USBD_CtlError(pdev, req);
          break;
      }
      break;

    default:
      USBD_CtlError(pdev, req);
      break;
  }

  return ret;
}

/**
  * @brief  USBD_StdEPReq
  *         Handle standard usb endpoint requests
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval status
  */
USBD_StatusTypeDef USBD_StdEPReq(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
  USBD_EndpointTypeDef *pep;
  uint8_t ep_addr;
  USBD_StatusTypeDef ret = USBD_OK;

  ep_addr = LOBYTE(req->wIndex);

  switch (req->bmRequest & USB_REQ_TYPE_MASK)
  {
    case USB_REQ_TYPE_CLASS:
    case USB_REQ_TYPE_VENDOR:
      /* Get the class index relative to this endpoint */
        pdev->classId = 0;
        /* Call the class data out function to manage the request */
        if (pdev->pClass->Setup != NULL)
        {
          ret = (USBD_StatusTypeDef)pdev->pClass->Setup(pdev, req);
        }
      break;

    case USB_REQ_TYPE_STANDARD:
      switch (req->bRequest)
      {
        case USB_REQ_SET_FEATURE:
          switch (pdev->dev_state)
          {
            case USBD_STATE_ADDRESSED:
              if ((ep_addr != 0x00U) && (ep_addr != 0x80U))
              {
                (void)USBD_LL_StallEP(pdev, ep_addr);
                (void)USBD_LL_StallEP(pdev, 0x80U);
              }
              else
              {
                USBD_CtlError(pdev, req);
              }
              break;

            case USBD_STATE_CONFIGURED:
              if (req->wValue == USB_FEATURE_EP_HALT)
              {
                if ((ep_addr != 0x00U) && (ep_addr != 0x80U) && (req->wLength == 0x00U))
                {
                  (void)USBD_LL_StallEP(pdev, ep_addr);
                }
              }
              (void)USBD_CtlSendStatus(pdev);

              break;

            default:
              USBD_CtlError(pdev, req);
              break;
          }
          break;

        case USB_REQ_CLEAR_FEATURE:

          switch (pdev->dev_state)
          {
            case USBD_STATE_ADDRESSED:
              if ((ep_addr != 0x00U) && (ep_addr != 0x80U))
              {
                (void)USBD_LL_StallEP(pdev, ep_addr);
                (void)USBD_LL_StallEP(pdev, 0x80U);
              }
              else
              {
                USBD_CtlError(pdev, req);
              }
              break;

            case USBD_STATE_CONFIGURED:
              if (req->wValue == USB_FEATURE_EP_HALT)
              {
                if ((ep_addr & 0x7FU) != 0x00U)
                {
                  (void)USBD_LL_ClearStallEP(pdev, ep_addr);
                }
                (void)USBD_CtlSendStatus(pdev);

                /* Get the class index relative to this interface */
                  pdev->classId = 0;
                  /* Call the class data out function to manage the request */
                  if (pdev->pClass->Setup != NULL)
                  {
                    ret = (USBD_StatusTypeDef)(pdev->pClass->Setup(pdev, req));
                  }
              }
              break;

            default:
              USBD_CtlError(pdev, req);
              break;
          }
          break;

        case USB_REQ_GET_STATUS:
          switch (pdev->dev_state)
          {
            case USBD_STATE_ADDRESSED:
              if ((ep_addr != 0x00U) && (ep_addr != 0x80U))
              {
                USBD_CtlError(pdev, req);
                break;
              }
              pep = ((ep_addr & 0x80U) == 0x80U) ? &pdev->ep_in[ep_addr & MAX_USBD_EP_IO_MASK] : \
                    &pdev->ep_out[ep_addr & MAX_USBD_EP_IO_MASK];

              pep->status = 0x0000U;

              (void)USBD_CtlSendData(pdev, (uint8_t *)&pep->status, 2U);
              break;

            case USBD_STATE_CONFIGURED:
              if ((ep_addr & 0x80U) == 0x80U)
              {
                if (pdev->ep_in[ep_addr & MAX_USBD_EP_IO_MASK].is_used == 0U)
                {
                  USBD_CtlError(pdev, req);
                  break;
                }
              }
              else
              {
                if (pdev->ep_out[ep_addr & MAX_USBD_EP_IO_MASK].is_used == 0U)
                {
                  USBD_CtlError(pdev, req);
                  break;
                }
              }

              pep = ((ep_addr & 0x80U) == 0x80U) ? &pdev->ep_in[ep_addr & MAX_USBD_EP_IO_MASK] : \
                    &pdev->ep_out[ep_addr & MAX_USBD_EP_IO_MASK];

              if ((ep_addr == 0x00U) || (ep_addr == 0x80U))
              {
                pep->status = 0x0000U;
              }
              else if (USBD_LL_IsStallEP(pdev, ep_addr) != 0U)
              {
                pep->status = 0x0001U;
              }
              else
              {
                pep->status = 0x0000U;
              }

              (void)USBD_CtlSendData(pdev, (uint8_t *)&pep->status, 2U);
              break;

            default:
              USBD_CtlError(pdev, req);
              break;
          }
          break;

        default:
          USBD_CtlError(pdev, req);
          break;
      }
      break;

    default:
      USBD_CtlError(pdev, req);
      break;
  }

  return ret;
}


/**
  * @brief  USBD_GetDescriptor
  *         Handle Get Descriptor requests
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval None
  */

typedef struct  {
	uint8_t bLength;
	uint8_t bDescriptorType;
	uint16_t wData[];
} __attribute__((packed)) str_desc;

#define USB_LANGID_ENGLISH_US 0x409


static uint8_t get_DescStr(USBD_HandleTypeDef *pdev, uint8_t no) {
  str_desc *sd;
  uint16_t idx;
  const uint8_t *cpc;

	if (no > MAX_DESC_STR_NUM) return USBD_FAIL;

	sd = (str_desc*)&pdev->ctl_buf;
	sd->bDescriptorType = USB_DESC_TYPE_STRING;

	if (no == 0) {
	    sd->bLength = 4;
	    sd->wData[0] = USB_LANGID_ENGLISH_US;
	} else {
	    cpc = (uint8_t*)pdev->DescStr[no-1];
	    if (!cpc) return USBD_FAIL;
	    idx = 0;
	    while (*cpc && (idx < (USB_MAX_CTL_BUF / 2))) {
		sd->wData[idx] = *cpc;
		idx++;
		cpc++;
	    }
	    sd->wData[idx] = 0;
	    sd->bLength = (idx * 2) + 2;
	}
	return USBD_OK;
}

static void USBD_GetDescriptor(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
  uint16_t len = 0U;
  uint8_t *pbuf = NULL;
  uint8_t err = 0U;
  str_desc *sd;

// printf("USBD_GetDescriptor 0x%04x  l=%d\r\n", req->wValue, req->wLength);

	switch (req->wValue >> 8) {

	case USB_DESC_TYPE_DEVICE:		//1
// printf("USBD_GetDescriptor l=0x%04x\r\n", pdev->DevDesc);
	  if (pdev->DevDesc) {
		pbuf = (uint8_t*)pdev->DevDesc;
		len = pdev->DevDesc->bLength;
	  } else {
	      USBD_CtlError(pdev, req);
	      err++;
	  }
//	  pbuf = pdev->pDesc->GetDeviceDescriptor(&len);
	  break;

	case USB_DESC_TYPE_CONFIGURATION:	//2
	    pbuf = (uint8_t *)pdev->pClass->GetFSConfigDescriptor(&len);
	    pbuf[1] = USB_DESC_TYPE_CONFIGURATION;
	  break;

	case USB_DESC_TYPE_STRING:		//3
#if 1
	    if (get_DescStr(pdev, req->wValue & 0x00ff) == USBD_OK) {
		pbuf = pdev->ctl_buf;
		sd = (str_desc*)&pdev->ctl_buf;
		len = sd->bLength;
	    } else {
		USBD_CtlError(pdev, req);
		err++;
	    }
#else
	    switch (req->wValue & 0x00ff) {
	    case USBD_IDX_LANGID_STR:		//0x0300
	      if (pdev->pDesc->GetLangIDStrDescriptor != NULL) {
		pbuf = pdev->pDesc->GetLangIDStrDescriptor(&len);
	      } else {
		USBD_CtlError(pdev, req);
		err++;
	      }
	      break;

	    case USBD_IDX_MFC_STR:		//0x0301
	      if (pdev->pDesc->GetManufacturerStrDescriptor != NULL) {
		pbuf = pdev->pDesc->GetManufacturerStrDescriptor(&len);
	      } else {
		USBD_CtlError(pdev, req);
		err++;
	      }
	      break;

	    case USBD_IDX_PRODUCT_STR:		//0x0302
	      if (pdev->pDesc->GetProductStrDescriptor != NULL) {
		pbuf = pdev->pDesc->GetProductStrDescriptor(&len);
	      } else {
		USBD_CtlError(pdev, req);
		err++;
	      }
	      break;

	    case USBD_IDX_SERIAL_STR:		//0x0303
	      if (pdev->pDesc->GetSerialStrDescriptor != NULL) {
		pbuf = pdev->pDesc->GetSerialStrDescriptor(&len);
	      } else {
		USBD_CtlError(pdev, req);
		err++;
	      }
	      break;

	    default:
	      USBD_CtlError(pdev, req);
	      err++;
	      break;
	    }
#endif
	    break;
/*
	case USB_DESC_TYPE_DEVICE_QUALIFIER:		//6
	    USBD_CtlError(pdev, req);
	    err++;
	    break;

	case USB_DESC_TYPE_OTHER_SPEED_CONFIGURATION:	//7
	    USBD_CtlError(pdev, req);
	    err++;
	    break;
*/
	default:
	  USBD_CtlError(pdev, req);
	  err++;
	  break;
	}
	if (err != 0U) return;

	if (req->wLength != 0U) {
	  if (len != 0U) {
	    len = MIN(len, req->wLength);
	    (void)USBD_CtlSendData(pdev, pbuf, len);
	  } else {
	    USBD_CtlError(pdev, req);
	  }
	} else {
	  (void)USBD_CtlSendStatus(pdev);
	}
}


/**
  * @brief  USBD_SetAddress
  *         Set device address
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval None
  */
static void USBD_SetAddress(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
  PCD_HandleTypeDef *hpcd;
  uint8_t dev_addr;

	if ((req->wIndex == 0U) && (req->wLength == 0U) && (req->wValue < 128U)) {
	  dev_addr = req->wValue;

	  if (pdev->dev_state == USBD_STATE_CONFIGURED) {
	    USBD_CtlError(pdev, req);

	  } else {
	    pdev->dev_address = dev_addr;
//	    (void)USBD_LL_SetUSBAddress(pdev, dev_addr);
	    hpcd = pdev->pData;
	    hpcd->USB_Address = dev_addr;
	    USB_DRD_FS->DADDR = USB_DADDR_EF;

	    (void)USBD_CtlSendStatus(pdev);

	    if (dev_addr != 0U) {
	      pdev->dev_state = USBD_STATE_ADDRESSED;
	    } else {
	      pdev->dev_state = USBD_STATE_DEFAULT;
	    }
	  }
	} else {
	  USBD_CtlError(pdev, req);
	}
}

/**
  * @brief  USBD_SetConfig
  *         Handle Set device configuration request
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval status
*/


static USBD_StatusTypeDef ClrClassConfig(USBD_HandleTypeDef *pdev, uint8_t cfgidx) {
  USBD_StatusTypeDef ret = USBD_OK;

	/* Clear configuration  and De-initialize the Class process */
	if (pdev->pClass->DeInit(pdev, cfgidx) != 0U) ret = USBD_FAIL;
	return ret;
}

static USBD_StatusTypeDef SetClassConfig(USBD_HandleTypeDef *pdev, uint8_t cfgidx) {
  USBD_StatusTypeDef ret = USBD_OK;

	if (pdev->pClass != NULL) {
	  /* Set configuration and Start the Class */
	    ret = (USBD_StatusTypeDef)pdev->pClass->Init(pdev, cfgidx);
	}
	return ret;
}

static USBD_StatusTypeDef USBD_SetConfig(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
  USBD_StatusTypeDef ret = USBD_OK;
  static uint8_t cfgidx;

  cfgidx = (uint8_t)(req->wValue);

  if (cfgidx > USBD_MAX_NUM_CONFIGURATION) {
    USBD_CtlError(pdev, req);
    return USBD_FAIL;
  }

  switch (pdev->dev_state) {
    case USBD_STATE_ADDRESSED:
      if (cfgidx != 0) {
        pdev->dev_config = cfgidx;
        ret = SetClassConfig(pdev, cfgidx);

        if (ret != USBD_OK) {
          USBD_CtlError(pdev, req);
          pdev->dev_state = USBD_STATE_ADDRESSED;

        } else {
          USBD_CtlSendStatus(pdev);
          pdev->dev_state = USBD_STATE_CONFIGURED;
        }
      } else {
        USBD_CtlSendStatus(pdev);
      }
      break;

    case USBD_STATE_CONFIGURED:
      if (cfgidx == 0) {
        pdev->dev_state = USBD_STATE_ADDRESSED;
        pdev->dev_config = cfgidx;
        (void)ClrClassConfig(pdev, cfgidx);
        (void)USBD_CtlSendStatus(pdev);

      } else if (cfgidx != pdev->dev_config) {
        /* Clear old configuration */
        (void)ClrClassConfig(pdev, (uint8_t)pdev->dev_config);
        /* set new configuration */
        pdev->dev_config = cfgidx;

        ret = SetClassConfig(pdev, cfgidx);
        if (ret != USBD_OK) {
          USBD_CtlError(pdev, req);
          (void)ClrClassConfig(pdev, (uint8_t)pdev->dev_config);
          pdev->dev_state = USBD_STATE_ADDRESSED;

        } else {
          (void)USBD_CtlSendStatus(pdev);
        }
      } else {
        (void)USBD_CtlSendStatus(pdev);
      }
      break;

    default:
      USBD_CtlError(pdev, req);
      (void)ClrClassConfig(pdev, cfgidx);
      ret = USBD_FAIL;
      break;
  }

  return ret;
}

/**
  * @brief  USBD_GetConfig
  *         Handle Get device configuration request
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval None
  */
static void USBD_GetConfig(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
	if (req->wLength != 1U) {
	    USBD_CtlError(pdev, req);
	} else {
	    switch (pdev->dev_state){
	    case USBD_STATE_DEFAULT:
	    case USBD_STATE_ADDRESSED:
	      pdev->dev_default_config = 0U;
	      (void)USBD_CtlSendData(pdev, (uint8_t *)&pdev->dev_default_config, 1U);
	      break;

	    case USBD_STATE_CONFIGURED:
	      (void)USBD_CtlSendData(pdev, (uint8_t *)&pdev->dev_config, 1U);
	      break;

	    default:
	      USBD_CtlError(pdev, req);
	      break;
	    }
	}
}

/**
  * @brief  USBD_GetStatus
  *         Handle Get Status request
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval None
  */
static void USBD_GetStatus(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
  uint16_t buf;
	buf = 0;
	USBD_CtlSendData(pdev, (uint8_t *)&buf, 2);
#if 0
 printf("USBD_GetStatus %d\r\n", pdev->dev_state);
  switch (pdev->dev_state)
  {
    case USBD_STATE_DEFAULT:
    case USBD_STATE_ADDRESSED:
    case USBD_STATE_CONFIGURED:
      if (req->wLength != 0x2U) {
        USBD_CtlError(pdev, req);
        break;
      }

#if (USBD_SELF_POWERED == 1U)
      pdev->dev_config_status = USB_CONFIG_SELF_POWERED;
#else
      pdev->dev_config_status = 0U;
#endif /* USBD_SELF_POWERED */

      if (pdev->dev_remote_wakeup != 0U) {
        pdev->dev_config_status |= USB_CONFIG_REMOTE_WAKEUP;
      }

      USBD_CtlSendData(pdev, (uint8_t *)&pdev->dev_config_status, 2U);
      break;

    default:
      USBD_CtlError(pdev, req);
      break;
  }
#endif
}


/**
  * @brief  USBD_SetFeature
  *         Handle Set device feature request
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval None
  */
static void USBD_SetFeature(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
  if (req->wValue == USB_FEATURE_REMOTE_WAKEUP)
  {
    pdev->dev_remote_wakeup = 1U;
    (void)USBD_CtlSendStatus(pdev);
  }
  else if (req->wValue == USB_FEATURE_TEST_MODE)
  {
    pdev->dev_test_mode = (uint8_t)(req->wIndex >> 8);
    (void)USBD_CtlSendStatus(pdev);
  }
  else
  {
    USBD_CtlError(pdev, req);
  }
}


/**
  * @brief  USBD_ClrFeature
  *         Handle clear device feature request
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval None
  */
static void USBD_ClrFeature(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
  switch (pdev->dev_state)
  {
    case USBD_STATE_DEFAULT:
    case USBD_STATE_ADDRESSED:
    case USBD_STATE_CONFIGURED:
      if (req->wValue == USB_FEATURE_REMOTE_WAKEUP)
      {
        pdev->dev_remote_wakeup = 0U;
        (void)USBD_CtlSendStatus(pdev);
      }
      break;

    default:
      USBD_CtlError(pdev, req);
      break;
  }
}


/**
  * @brief  USBD_ParseSetupRequest
  *         Copy buffer into setup structure
  * @param  req: usb request
  * @param  pdata: setup data pointer
  * @retval None
  */
void USBD_ParseSetupRequest(USBD_SetupReqTypedef *req, uint8_t *pdata)
{
  uint8_t *pbuff = pdata;

  req->bmRequest = *(uint8_t *)(pbuff);

  pbuff++;
  req->bRequest = *(uint8_t *)(pbuff);

  pbuff++;
  req->wValue = SWAPBYTE(pbuff);

  pbuff++;
  pbuff++;
  req->wIndex = SWAPBYTE(pbuff);

  pbuff++;
  pbuff++;
  req->wLength = SWAPBYTE(pbuff);
}


/**
  * @brief  USBD_CtlError
  *         Handle USB low level Error
  * @param  pdev: device instance
  * @param  req: usb request
  * @retval None
  */
void USBD_CtlError(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req)
{
  UNUSED(req);

  (void)USBD_LL_StallEP(pdev, 0x80U);
  (void)USBD_LL_StallEP(pdev, 0U);
}


/**
  * @brief  USBD_GetString
  *         Convert Ascii string into unicode one
  * @param  desc : descriptor buffer
  * @param  unicode : Formatted string buffer (unicode)
  * @param  len : descriptor length
  * @retval None
  */
void USBD_GetString(uint8_t *desc, uint8_t *unicode, uint16_t *len)
{
  uint8_t idx = 0U;
  uint8_t *pdesc;

  if (desc == NULL)
  {
    return;
  }

  pdesc = desc;
  *len = MIN(USBD_MAX_STR_DESC_SIZ, ((uint16_t)USBD_GetLen(pdesc) * 2U) + 2U);

  unicode[idx] = *(uint8_t *)len;
  idx++;
  unicode[idx] = USB_DESC_TYPE_STRING;
  idx++;

  while (*pdesc != (uint8_t)'\0')
  {
    unicode[idx] = *pdesc;
    pdesc++;
    idx++;

    unicode[idx] = 0U;
    idx++;
  }
}


/**
  * @brief  USBD_GetLen
  *         return the string length
   * @param  buf : pointer to the ascii string buffer
  * @retval string length
  */
static uint8_t USBD_GetLen(uint8_t *buf)
{
  uint8_t  len = 0U;
  uint8_t *pbuff = buf;

  while (*pbuff != (uint8_t)'\0')
  {
    len++;
    pbuff++;
  }

  return len;
}
/**
  * @}
  */


/**
  * @}
  */


/**
  * @}
  */

