/**
  ******************************************************************************
  * @file    usbd_tmc.c
  * @attention
  *
  * Copyright (c) 2015 STMicroelectronics.
  * All rights reserved.
  *
  * This software is licensed under terms that can be found in the LICENSE file
  * in the root directory of this software component.
  * If no LICENSE file comes with this software, it is provided AS-IS.
  *
  ******************************************************************************
*/

#include "main.h"
#include "usbd_tmc_interface.h"
#include "usbd_ctlreq.h"
#include "usbd_core.h"

#define USB_VID				0x0483
#define USB_PID				0x5740
#define USB_LANGID_STRING		0x409
#define USB_MANUFACTURER_STRING		"STMicroelectronics"
#define USB_PRODUCT_STRING		"STM32 Virtual ComPort in FS Mode"
//#define USB_CONFIGURATION_STRING	"VCP Config"
//#define USB_INTERFACE_STRING		"VCP Interface"
#ifdef INIMP_2K5
#define USB_SERIAL_STRING		LAY_SNR".00 2k5"
#elif defined(INIMP_10K)
#define USB_SERIAL_STRING		LAY_SNR".00 10k"
#else
#define USB_SERIAL_STRING		LAY_SNR".00 50"
#endif

static uint8_t USB_TMC_Init(USBD_HandleTypeDef *pdev, uint8_t cfgidx);
static uint8_t USB_TMC_DeInit(USBD_HandleTypeDef *pdev, uint8_t cfgidx);
static uint8_t USB_TMC_Setup(USBD_HandleTypeDef *pdev, USBD_SetupReqTypedef *req);
static uint8_t USB_TMC_DataIn(USBD_HandleTypeDef *pdev, uint8_t epnum);
static uint8_t USB_TMC_DataOut(USBD_HandleTypeDef *pdev, uint8_t epnum);
static uint8_t USB_TMC_EP0_RxReady(USBD_HandleTypeDef *pdev);
static const uint8_t *TMC_GetFSCfgDesc(uint16_t *length);

/* CDC interface class callbacks structure */
USBD_ClassTypeDef const USBD_CDC =
{
  USB_TMC_Init,
  USB_TMC_DeInit,
  USB_TMC_Setup,
  NULL,                 /* EP0_TxSent */
  USB_TMC_EP0_RxReady,
  USB_TMC_DataIn,
  USB_TMC_DataOut,
  NULL,
  NULL,
  NULL,
  TMC_GetFSCfgDesc,
  NULL,
};

typedef struct {
	uint8_t bLength;
	uint8_t bDescriptorType;
	uint8_t bInterfaceNumber;
	uint8_t bAlternateSetting;
	uint8_t bNumEndpoints;
	uint8_t bInterfaceClass;
	uint8_t bInterfaceSubClass;
	uint8_t bInterfaceProtocol;
	uint8_t iInterface;
} __attribute__((__packed__)) USBD_InterfaceDesc;

struct Ttmc_CfgDesc {
	USBD_ConfigDescTypeDef config;
	USBD_InterfaceDesc ifdesc;
	USBD_EpDescTypeDef ep_out;
	USBD_EpDescTypeDef ep_in;
} __attribute__((__packed__));

/* USB Standard Device Descriptor */
static const USB_DeviceDescriptor DevDesc = {
	.bLength = sizeof(USB_DeviceDescriptor),
	.bDescriptorType = USB_DESC_TYPE_DEVICE,
	.bcdUSB = 0x0200,
	.bDeviceClass = 0, // on interface
	.bDeviceSubClass = 0,
	.bDeviceProtocol = 0,
	.bMaxPacketSize0 = 64,
	.idVendor = USB_VID,
	.idProduct = USB_PID,
	.bcdDevice = 0x0200,
	.iManufacturer = 1,
	.iProduct = 2,
	.iSerialNumber = 3,
	.bNumConfigurations = 1,
};

static const struct Ttmc_CfgDesc tmc_CfgDesc =
{
	.config.bLength = sizeof(USBD_ConfigDescTypeDef),
	.config.bDescriptorType = USB_DESC_TYPE_CONFIGURATION,
	.config.wTotalLength = sizeof(struct Ttmc_CfgDesc),
	.config.bNumInterfaces = 1,
	.config.bConfigurationValue = 1,
	.config.iConfiguration = 0,
	.config.bmAttributes = 0x80,
	.config.bMaxPower = 50,

	.ifdesc.bLength = sizeof(USBD_InterfaceDesc),
	.ifdesc.bDescriptorType = USB_DESC_TYPE_INTERFACE,
	.ifdesc.bInterfaceNumber = 0,
	.ifdesc.bAlternateSetting = 0,
	.ifdesc.bNumEndpoints = 2,
	.ifdesc.bInterfaceClass = USB_CLASS_APPLICATION,
	.ifdesc.bInterfaceSubClass = USB_APPLICATION_SUBCLASS_TMC,
	.ifdesc.bInterfaceProtocol = USB_TMC_PROTOCOL_NONE,
	.ifdesc.iInterface = 0,

	.ep_out.bLength = sizeof(USBD_EpDescTypeDef),
	.ep_out.bDescriptorType = USB_DESC_TYPE_ENDPOINT,
	.ep_out.bEndpointAddress = TMC_OUT_EP,
	.ep_out.bmAttributes = USBD_EP_TYPE_BULK,
	.ep_out.wMaxPacketSize = 64,
	.ep_out.bInterval = 1,

	.ep_in.bLength = sizeof(USBD_EpDescTypeDef),
	.ep_in.bDescriptorType = USB_DESC_TYPE_ENDPOINT,
	.ep_in.bEndpointAddress = TMC_IN_EP,
	.ep_in.bmAttributes = USBD_EP_TYPE_BULK,
	.ep_in.wMaxPacketSize = 64,
	.ep_in.bInterval = 1,
};

#define CDCInEpAdd	CDC_IN_EP
#define CDCOutEpAdd	CDC_OUT_EP

USB_TMC_HandleTypeDef tmc_h;

/**
  * @brief  USB_TMC_Init
  *         Initialize the CDC interface
  * @param  pdev: device instance
  * @param  cfgidx: Configuration index
  * @retval status
  */
static uint8_t USB_TMC_Init(USBD_HandleTypeDef *pdev, uint8_t cfgidx)
{
  UNUSED(cfgidx);

//  tmc_h = (USB_TMC_HandleTypeDef *)USBD_malloc(sizeof(USB_TMC_HandleTypeDef));

//  if (tmc_h == NULL)
//  {
//    pdev->pClassDataCmsit = NULL;
//    return (uint8_t)USBD_EMEM;
//  }

  USBD_memset(&tmc_h, 0, sizeof(USB_TMC_HandleTypeDef));

  pdev->pClassDataCmsit = (void *)&tmc_h;
  pdev->pClassData = pdev->pClassDataCmsit;

    /* Open EP IN */
    (void)USBD_LL_OpenEP(pdev, CDCInEpAdd, USBD_EP_TYPE_BULK,
                         CDC_DATA_FS_IN_PACKET_SIZE);

    pdev->ep_in[CDCInEpAdd & MAX_USBD_EP_IO_MASK].is_used = 1U;

    /* Open EP OUT */
    (void)USBD_LL_OpenEP(pdev, CDCOutEpAdd, USBD_EP_TYPE_BULK,
                         CDC_DATA_FS_OUT_PACKET_SIZE);

    pdev->ep_out[CDCOutEpAdd & MAX_USBD_EP_IO_MASK].is_used = 1U;

//    /* Set bInterval for CMD Endpoint */
//    pdev->ep_in[CDCCmdEpAdd & 0xFU].bInterval = CDC_FS_BINTERVAL;

  /* Open Command IN EP */
//  (void)USBD_LL_OpenEP(pdev, CDCCmdEpAdd, USBD_EP_TYPE_INTR, CDC_CMD_PACKET_SIZE);
//  pdev->ep_in[CDCCmdEpAdd & MAX_USBD_EP_IO_MASK].is_used = 1U;

//  tmc_h.RxBuffer = NULL;

//	USB_TMC_MemInit();
	USB_TMC_SetTxBuffer(pdev, TmcTxBuffer, 0);
//	USB_TMC_SetRxBuffer(pdev, TmcRxBuffer);

  /* Init Xfer states */
  tmc_h.TxState = 0;
  tmc_h.RxState = 0;

//  if (tmc_h.RxBuffer == NULL)
//  {
//    return (uint8_t)USBD_EMEM;
//  }

    /* Prepare Out endpoint to receive next packet */
    (void)USBD_LL_PrepareReceive(pdev, CDCOutEpAdd, TmcRxBuffer,
                                 CDC_DATA_FS_OUT_PACKET_SIZE);

	USB_TmcDescInit(pdev);
  return (uint8_t)USBD_OK;
}

/**
  * @brief  USB_TMC_Init
  *         DeInitialize the CDC layer
  * @param  pdev: device instance
  * @param  cfgidx: Configuration index
  * @retval status
  */
static uint8_t USB_TMC_DeInit(USBD_HandleTypeDef *pdev, uint8_t cfgidx)
{
  UNUSED(cfgidx);


  /* Close EP IN */
  (void)USBD_LL_CloseEP(pdev, CDCInEpAdd);
  pdev->ep_in[CDCInEpAdd & MAX_USBD_EP_IO_MASK].is_used = 0U;

  /* Close EP OUT */
  (void)USBD_LL_CloseEP(pdev, CDCOutEpAdd);
  pdev->ep_out[CDCOutEpAdd & MAX_USBD_EP_IO_MASK].is_used = 0U;

  /* Close Command IN EP */
//  (void)USBD_LL_CloseEP(pdev, CDCCmdEpAdd);
//  pdev->ep_in[CDCCmdEpAdd & MAX_USBD_EP_IO_MASK].is_used = 0U;
//  pdev->ep_in[CDCCmdEpAdd & MAX_USBD_EP_IO_MASK].bInterval = 0U;

  /* DeInit  physical Interface components */
//  if (pdev->pClassDataCmsit != NULL)
//  {
//    ((USB_TMC_ItfTypeDef *)pdev->pUserData)->DeInit();
//    (void)USBD_free(pdev->pClassDataCmsit);
    pdev->pClassDataCmsit = NULL;
    pdev->pClassData = NULL;
//  }

  return (uint8_t)USBD_OK;
}

static const struct usb_tmc_get_capabilities_response capabilities = {
	.USBTMC_status = USB_TMC_STATUS_SUCCESS,
	.bcdUSBTMC = 0x0100,
	.reserved0 = 0,
	.reserved1 = {0},
	.reserved_subclass = {0},
	.interface_capabilities = USB_TMC_INTERFACE_CAPABILITY_INDICATOR_PULSE,
	.device_capabilities = 0
};

static void USB_TMC_Control (uint8_t cmd, uint8_t* pbuf, uint16_t length) { 

	switch (cmd) {
	case USB_TMC_REQ_INITIATE_CLEAR:
// printf("CDC_Control USB_TMC_REQ_INITIATE_CLEAR l=%d c=0x%02x\r\n", length, pbuf[0]);
	  tmc_h.RxState = 0;
	  tmc_h.RxOffset = 0;
	  break;

	case USB_TMC_REQ_CHECK_CLEAR_STATUS:
// printf("CDC_Control USB_TMC_REQ_CHECK_CLEAR_STATUS l=%d c=0x%02x 0x%02x\r\n", length, pbuf[0], pbuf[1]);
	  pbuf[0] = USB_TMC_STATUS_SUCCESS;
	  pbuf[1] = 0;
	  break;

	case USB_TMC_REQ_GET_CAPABILITIES:
      //	printf("USB_TMC_REQ_GET_CAPABILITIES l=%d\r\n", length);
	  memcpy(pbuf, &capabilities, sizeof(capabilities));
	  break;
    
	default:
#if DEBUG_CONSOLE_ON
 printf("CDC_Control cmd=%d l=%d\r\n", cmd, length);
#endif
	  break;
	}
}


/**
  * @brief  USB_TMC_Setup
  *         Handle the CDC specific requests
  * @param  pdev: instance
  * @param  req: usb requests
  * @retval status
  */
static uint8_t USB_TMC_Setup(USBD_HandleTypeDef *pdev,
                              USBD_SetupReqTypedef *req)
{
//  USB_TMC_HandleTypeDef *tmc_h = (USB_TMC_HandleTypeDef *)pdev->pClassDataCmsit;
  uint16_t len;
  uint8_t ifalt = 0U;
  uint16_t status_info = 0U;

	switch (req->bmRequest & USB_REQ_TYPE_MASK) {
	case USB_REQ_TYPE_CLASS:
// printf("CDC_REQ_MAX_DATA_SIZE req=0x%02x  l=%d\r\n", req->bmRequest, req->wLength);
	    if (req->wLength != 0U) {
	      if ((req->bmRequest & 0x80U) != 0U) {
		USB_TMC_Control(req->bRequest, (uint8_t *)tmc_h.data, req->wLength);
		len = MIN(CDC_REQ_MAX_DATA_SIZE, req->wLength);
		USBD_CtlSendData(pdev, (uint8_t *)tmc_h.data, len);
	      } else {
		tmc_h.CmdOpCode = req->bRequest;
		tmc_h.CmdLength = (uint8_t)MIN(req->wLength, USB_MAX_EP0_SIZE);

		USBD_CtlPrepareRx(pdev, (uint8_t *)tmc_h.data, tmc_h.CmdLength);
	      }
	    } else {
		USB_TMC_Control(req->bRequest, (uint8_t*)req, 0);
	    }
	    break;

	case USB_REQ_TYPE_STANDARD:

	    switch (req->bRequest) {
	    case USB_REQ_GET_STATUS:
		if (pdev->dev_state == USBD_STATE_CONFIGURED) {
		  (void)USBD_CtlSendData(pdev, (uint8_t *)&status_info, 2U);
		} else goto L_error;
		break;

	    case USB_REQ_GET_INTERFACE:
		if (pdev->dev_state == USBD_STATE_CONFIGURED) {
		  (void)USBD_CtlSendData(pdev, &ifalt, 1U);
		} else goto L_error;
		break;

	    case USB_REQ_SET_INTERFACE:
		if (pdev->dev_state != USBD_STATE_CONFIGURED) goto L_error;
		break;

	    case USB_REQ_CLEAR_FEATURE:
		break;

	    default:
		goto L_error;
	    }
	    break;

	default:
	  goto L_error;
	}

	return USBD_OK;

L_error:
	USBD_CtlError(pdev, req);
	return USBD_FAIL;
}

/**
  * @brief  USB_TMC_DataIn
  *         Data sent on non-control IN endpoint
  * @param  pdev: device instance
  * @param  epnum: endpoint number
  * @retval status
  */
static uint8_t USB_TMC_DataIn(USBD_HandleTypeDef *pdev, uint8_t epnum)
{
//  USB_TMC_HandleTypeDef *tmc_h;
  PCD_HandleTypeDef *hpcd = (PCD_HandleTypeDef *)pdev->pData;

//  if (pdev->pClassDataCmsit == NULL)
//  {
//    return (uint8_t)USBD_FAIL;
 // }

//  tmc_h = (USB_TMC_HandleTypeDef *)pdev->pClassDataCmsit;

  if ((pdev->ep_in[epnum & MAX_USBD_EP_IO_MASK].total_length > 0U) &&
      ((pdev->ep_in[epnum & MAX_USBD_EP_IO_MASK].total_length % hpcd->IN_ep[epnum & MAX_USBD_EP_IO_MASK].maxpacket) == 0U))
  {
    /* Update the packet total length */
    pdev->ep_in[epnum & MAX_USBD_EP_IO_MASK].total_length = 0U;

    /* Send ZLP */
    (void)USBD_LL_Transmit(pdev, epnum, NULL, 0U);
  }
  else
  {
    tmc_h.TxState = 0U;
    USB_TMC_TransmitCplt(tmc_h.TxBuffer, &tmc_h.TxLength, epnum);
  }

  return (uint8_t)USBD_OK;
}

/**
  * @brief  USB_TMC_DataOut
  *         Data received on non-control Out endpoint
  * @param  pdev: device instance
  * @param  epnum: endpoint number
  * @retval status
  */
static uint8_t USB_TMC_DataOut(USBD_HandleTypeDef *pdev, uint8_t epnum)
{
  uint32_t RxLength;

  /* Get the received data length */
	RxLength = USBD_LL_GetRxDataSize(pdev, epnum);

	USB_TMC_Receive(pdev, RxLength);
	return (uint8_t)USBD_OK;
}

/**
  * @brief  USB_TMC_EP0_RxReady
  *         Handle EP0 Rx Ready event
  * @param  pdev: device instance
  * @retval status
  */
static uint8_t USB_TMC_EP0_RxReady(USBD_HandleTypeDef *pdev) {
// printf("USB_TMC_EP0_RxReady\r\n");
	if ((pdev->pUserData != NULL) && (tmc_h.CmdOpCode != 0xFF)) {
	  USB_TMC_Control(tmc_h.CmdOpCode, (uint8_t *)tmc_h.data, (uint16_t)tmc_h.CmdLength);
	  tmc_h.CmdOpCode = 0xFF;
	}
	return (uint8_t)USBD_OK;
}

/**
  * @brief  TMC_GetFSCfgDesc
  *         Return configuration descriptor
  * @param  length : pointer data length
  * @retval pointer to descriptor buffer
  */
static const uint8_t *TMC_GetFSCfgDesc(uint16_t *length) {
	*length = (uint16_t)sizeof(tmc_CfgDesc);
	return (uint8_t*)&tmc_CfgDesc;
}

/**
  * @brief  USB_TMC_SetTxBuffer
  * @param  pdev: device instance
  * @param  pbuff: Tx Buffer
  * @param  length: length of data to be sent
  * @param  ClassId: The Class ID
  * @retval status
  */
uint8_t USB_TMC_SetTxBuffer(USBD_HandleTypeDef *pdev,
                             uint8_t *pbuff, uint32_t length)
{
//  USB_TMC_HandleTypeDef *tmc_h = (USB_TMC_HandleTypeDef *)pdev->pClassDataCmsit;

//  if (tmc_h == NULL)
//  {
//    return (uint8_t)USBD_FAIL;
//  }

  tmc_h.TxBuffer = pbuff;
  tmc_h.TxLength = length;

  return (uint8_t)USBD_OK;
}

#if 0
/**
  * @brief  USB_TMC_SetRxBuffer
  * @param  pdev: device instance
  * @param  pbuff: Rx Buffer
  * @retval status
  */
uint8_t USB_TMC_SetRxBuffer(USBD_HandleTypeDef *pdev, uint8_t *pbuff)
{
//  USB_TMC_HandleTypeDef *tmc_h = (USB_TMC_HandleTypeDef *)pdev->pClassDataCmsit;

//  if (tmc_h == NULL)
//  {
//    return (uint8_t)USBD_FAIL;
//  }

  tmc_h.RxBuffer = pbuff;

  return (uint8_t)USBD_OK;
}
#endif

/**
  * @brief  USB_TMC_TransmitPacket
  *         Transmit packet on IN endpoint
  * @param  pdev: device instance
  * @param  ClassId: The Class ID
  * @retval status
  */
uint8_t USB_TMC_TransmitPacket(USBD_HandleTypeDef *pdev)
{
//  USB_TMC_HandleTypeDef *tmc_h = (USB_TMC_HandleTypeDef *)pdev->pClassDataCmsit;

  USBD_StatusTypeDef ret = USBD_BUSY;

//  if (tmc_h == NULL)
//  {
//    return (uint8_t)USBD_FAIL;
//  }

  if (tmc_h.TxState == 0U)
  {
    /* Tx Transfer in progress */
    tmc_h.TxState = 1U;

    /* Update the packet total length */
//    pdev->ep_in[CDCInEpAdd & MAX_USBD_EP_IO_MASK].total_length = tmc_h.TxLength;

    /* Transmit next packet */
    (void)USBD_LL_Transmit(pdev, CDCInEpAdd, tmc_h.TxBuffer, tmc_h.TxLength);

// printf("Txd l=%ld\r\n", tmc_h.TxLength);

    ret = USBD_OK;
  }

  return (uint8_t)ret;
}

/**
  * @brief  USB_TMC_ReceivePacket
  *         prepare OUT Endpoint for reception
  * @param  pdev: device instance
  * @retval status
  */
uint8_t USB_TMC_ReceivePacket(USBD_HandleTypeDef *pdev)
{
//  USB_TMC_HandleTypeDef *tmc_h = (USB_TMC_HandleTypeDef *)pdev->pClassDataCmsit;

//  if (!tmc_h) return (uint8_t)USBD_FAIL;

    /* Prepare Out endpoint to receive next packet */
    (void)USBD_LL_PrepareReceive(pdev, CDCOutEpAdd, TmcRxBuffer,
                                 CDC_DATA_FS_OUT_PACKET_SIZE);

  return (uint8_t)USBD_OK;
}

//__weak void USB_TMC_MemInit() {
//}

//__weak void USB_TMC_Control (uint8_t cmd, uint8_t* pbuf, uint16_t length) {
//}

__weak void USB_TMC_TransmitCplt(uint8_t *Buf, uint32_t *Len, uint8_t epnum) {
}

//__weak void USB_TMC_Receive(uint8_t* Buf, uint32_t *Len) {
//}

void USB_TmcDescInit(USBD_HandleTypeDef *pdev) {
	pdev->DescStr[0] = USB_MANUFACTURER_STRING;
	pdev->DescStr[1] = USB_PRODUCT_STRING;
	pdev->DescStr[2] = USB_SERIAL_STRING;
	pdev->DescStr[3] = NULL;
	pdev->DescStr[4] = NULL;
	pdev->DevDesc = &DevDesc;
}
